#include <stdLoc.h>
/* here is a fast floating point to ascii converter we developed for the
** BEVALAC control system for storing control system state.
**
** This work done by Loren Shalz, Alan Biocca, and Maurice McEvoy.
**
**
** ftoa converts float to decimal ascii using the MC68881/2 Floating_Point
** Coprocessor.  It utilizes a single line of assembly language to access
** the FPU's floating point to packed BCD.  The remainder of the code is
** written in C.  The WRS 5.0.1 release gnu cross tools were used (based
** on gcc 1.37).  Your mileage may vary -- TEST THESE ROUTINES THOROUGHLY.
**
** ftoa must be compiled with the gnu compiler because cc's assembler
** does not recoginze the k factor(precision) in the inline fmovep command.
**           cc68k -c -o ftoa.vx ftoa.c
**
** If any code before the fmovep is changed be sure to get the assembler
** output and verify that value has been put into fp0, that a3 still
** has the address of packed_bcd and that d4 still has precision.
** Adjust fmovep appropriately if any changes are observed.
**
** See the MC68881 Floating-Point Coprocessor Users Manual for the
** packed decimal real format produced by the fmovep. The following
** defines aid in cracking that format to produce the decimal ascii.
*/

#define MANTISSA_SIGN ( ( unsigned char ) ( 1 << 7 ) )
#define EXPONENT_SIGN ( ( unsigned char ) ( 1 << 6 ) )
#define NAN_BITS      ( ( unsigned char ) ( 3 << 4 ) )
#define ASCII_ZERO    ( ( unsigned char ) 0x30 )
#define RIGHT_NIBBLE  ( ( unsigned char ) 0xf )
#define NIBBLE_SIZE   ( ( unsigned char ) 0x4 )
#define PACKED_BCD_SIZE  12

int
ftoa( value, precision, width, string )

float value;		/* value to be converted to decimal ascii */

int precision;          /* 1 <= precision(no. significant digits) <= 17 */

int width;              /* minimum size of returned string.  String is left */
			/* adjusted with blank fill if necessary */

char *string;		/* pointer to string to receive converted value */
                        /* format returned is [-]d.dddddddddddddddde[-]ddd */
			/* so string should be sized to hold at least the
			/* larger of width or  precision + 7 characters */
			/* plus a terminating null. Trailing zeros in the */
			/* mantissa and zero exponents are dropped */

/* ftoa returns the length of string */

{
    unsigned char packed_bcd[PACKED_BCD_SIZE];

    register float v = value;
    register int p = precision;
    register int w = width;
    register char *s = string;
    register unsigned char *c = &packed_bcd[0];
    int i;
    int trailing_zeros;
    int trailing_blanks;
    int right_adjuster;
    unsigned char next_digit;

    asm( " fmovep fp2,a3@{d4}" );	/* convert to packed bcd. */

    if ( precision < 1 ) precision = 1;
    if ( precision > 17 ) precision = 17;

    if ( ( c[0] & NAN_BITS ) != 0 )
    {
	strcpy( s, "NAN" );
	return;
    }

/* unpack mantissa--always return at least 1 digit and '.' */

    if ( ( c[0] & MANTISSA_SIGN ) != 0 ) *s++ = '-';	/* mantissa sign */
    *s++ = ( c[3] & RIGHT_NIBBLE ) + ASCII_ZERO;	/* most sign. digit */
    *s++ = '.';						/* decimal point */

    trailing_zeros = 0;
    --precision;			/* 1st digit unpacked above */
    right_adjuster = NIBBLE_SIZE;
    c = &packed_bcd[4];
    for ( i = 0; i < precision ; ++i )	/* unpack rest of digits */
    {
	 next_digit = ( c[i/2] >> right_adjuster ) & RIGHT_NIBBLE;

	if ( next_digit == '\0' )
	    ++trailing_zeros;
	else
	    trailing_zeros = 0;

	*s++ = next_digit + ASCII_ZERO;

	if ( right_adjuster == 0 ) 
	    right_adjuster = NIBBLE_SIZE;
	else
	    right_adjuster = 0;
    }
    s = s - trailing_zeros;

/* unpack exponent--drop it if it is zero.  */

    c = &packed_bcd[0];
    if ( ( ( next_digit = ( c[0] & RIGHT_NIBBLE ) ) + c[1] ) != 0 )
    {
        *s++ = 'e';					/* nonzero exponent */
        if ( ( c[0] & EXPONENT_SIGN ) != 0 ) *s++ = '-';

        if ( next_digit  != 0 )
        {
            *s++ = next_digit + ASCII_ZERO;		/* 3 digit exponent */
            *s++ = ( c[1] >> NIBBLE_SIZE ) + ASCII_ZERO;
        }
        else if ( ( next_digit = ( c[1] >> NIBBLE_SIZE ) ) != 0 )
            *s++ = next_digit + ASCII_ZERO;		/* 2 digit exponent */

        *s++ = ( c[1] & RIGHT_NIBBLE ) + ASCII_ZERO;	/* 1 digit exponent */
    }

    if ( ( trailing_blanks = width - ( s - string ) ) > 0 ) /* blank pad */
        for ( i = 0; i < trailing_blanks; ++i )
	    *s++ = ' ';
    *s = '\0';
    return s - string;
}






int
ftoa2( value, precision, width, string )

float value;		/* value to be converted to decimal ascii */

int precision;          /* 1 <= precision(no. significant digits) <= 17 */

int width;              /* minimum size of returned string.  String is left */
			/* adjusted with blank fill if necessary */

char *string;		/* pointer to string to receive converted value */
                        /* format returned is [-]ddddd.ddddd
			/* so string should be sized to hold at least the
			/* larger of width or  precision + 7 characters */
			/* plus a terminating null. Trailing zeros in the */
			/* mantissa and zero exponents are dropped */

/* ftoa returns the length of string */

{
    unsigned char packed_bcd[PACKED_BCD_SIZE];

    register float v = value;
    register int p = precision;
    register int w = width;
    register char *s = string;
    register unsigned char *c = &packed_bcd[0];
    int i;
    int trailing_zeros;
    int trailing_blanks;
    int right_adjuster;
    unsigned char next_digit;
    int doneDot = 0;
    int exp;

    asm( " fmovep fp2,a3@{d4}" );	/* convert to packed bcd. */

    if ( precision < 1 ) precision = 1;
    if ( precision > 17 ) precision = 17;

    if ( ( c[0] & NAN_BITS ) != 0 )
    {
	strcpy( s, "NAN" );
	return;
    }

/* decode exponent - to determine how to represent mantissa */

    c = &packed_bcd[0];
    exp = 0;
    if ( ( ( next_digit = ( c[0] & RIGHT_NIBBLE ) ) + c[1] ) != 0 )
    {
        /* nonzero exponent */

        if ( next_digit  != 0 )
        {
	    /* 3 digit exponent */
	    return ftoa(value, precision, width, string);
        }
        else if ( ( next_digit = ( c[1] >> NIBBLE_SIZE ) ) != 0 )
	{
	    /* 2 digit exponent */
	    return ftoa(value, precision, width, string);
	}

        exp = ( c[1] & RIGHT_NIBBLE ) ;
        if ( ( c[0] & EXPONENT_SIGN ) != 0 ) exp = -exp;
    }
    if ( ( c[0] & MANTISSA_SIGN ) != 0 ) *s++ = '-';	/* mantissa sign */
/* unpack mantissa--always return at least 1 digit and '.' */
    if (exp >= 0)
    {
	fprintf(stderr, "exp = %d\n", exp);

	*s++ = ( c[3] & RIGHT_NIBBLE ) + ASCII_ZERO;	/* most sign. digit */
	trailing_zeros = 0;
	--precision;			/* 1st digit unpacked above */
	right_adjuster = NIBBLE_SIZE;
	c = &packed_bcd[4];

	for ( i = 0; i < precision ; ++i )	/* unpack rest of digits */
	{
	    if (exp == 0) {
		*s++ = '.';			/* decimal point */
		doneDot = 1;
	    }
	    --exp;
	     next_digit = ( c[i/2] >> right_adjuster ) & RIGHT_NIBBLE;

	    if ( doneDot && next_digit == '\0' )
		++trailing_zeros;
	    else
		trailing_zeros = 0;

	    *s++ = next_digit + ASCII_ZERO;

	    if ( right_adjuster == 0 ) 
		right_adjuster = NIBBLE_SIZE;
	    else
		right_adjuster = 0;
	}
	if (exp == 0)
	{
	    if (!doneDot)
	    {
		*s++ = '.';			/* decimal point */
		doneDot = 1;
	    }
	} else {
	    fprintf(stderr, "exp = %d, out of precision\n", exp);
	}
    } else {
	return ftoa(value, precision, width, string);
	/* exponent less than zero */
    }
    s = s - trailing_zeros;


    if ( ( trailing_blanks = width - ( s - string ) ) > 0 ) /* blank pad */
        for ( i = 0; i < trailing_blanks; ++i )
	    *s++ = ' ';
    *s = '\0';
    fprintf(stderr, "returning %s\n", string);
    return s - string;
}
